
// I2Cdev and MPU6050 must be installed as libraries, or else the .cpp/.h files
// for both classes must be in the include path of your project
#include "I2Cdev.h"

#include "MPU6050_6Axis_MotionApps20.h" 
/*
!!IMPORTANT!! MPU6050_DMP_FIFO_RATE_DIVISOR must be set to at least 0x06 in 
libraries/MPU6050/MPU6050_6Axis_MotionApps20.h
*/
#include "MPU6050.h" // not necessary if using MotionApps include file

// Arduino Wire library is required if I2Cdev I2CDEV_ARDUINO_WIRE implementation
// is used in I2Cdev.h
#if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
#include "Wire.h"
#endif
#include <Servo.h>
MPU6050 mpu;

// Define the 3 servo motors
Servo servo0;
Servo servo1;
Servo servo2;
float correct;
int j = 0;

#define OUTPUT_READABLE_YAWPITCHROLL

#define INTERRUPT_PIN 2    // use pin 2 on Arduino Uno & most boards

bool blinkState = false;

// MPU control/status vars
bool dmpReady = false;    // set true if DMP init was successful
uint8_t mpuIntStatus;     // holds actual interrupt status byte from MPU
uint8_t devStatus;            // return status after each device operation (0 = success, !0 = error)
uint16_t packetSize;        // expected DMP packet size (default is 42 bytes)
uint16_t fifoCount;         // count of all bytes currently in FIFO
uint8_t fifoBuffer[64]; // FIFO storage buffer

// orientation/motion vars
Quaternion q;                     // [w, x, y, z]                 quaternion container
VectorInt16 aa;                 // [x, y, z]                        accel sensor measurements
VectorInt16 aaReal;         // [x, y, z]                        gravity-free accel sensor measurements
VectorInt16 aaWorld;        // [x, y, z]                        world-frame accel sensor measurements
VectorFloat gravity;        // [x, y, z]                        gravity vector
float euler[3];                 // [psi, theta, phi]        Euler angle container
float ypr[3];                     // [yaw, pitch, roll]     yaw/pitch/roll container and gravity vector

// packet structure for InvenSense teapot demo
// ================================================================
// ===                             INTERRUPT DETECTION ROUTINE                                ===
// ================================================================

volatile bool mpuInterrupt = false;         // indicates whether MPU interrupt pin has gone high
void dmpDataReady() {
    mpuInterrupt = true;
}

// how much to scale the outputs by. If negative, the servo moves backwards.
float corrections[3] = {3.0, -3.0, 3.0};

// used to zero the yaw, pitch and roll of the gimbal
float offset[3] = {0.0,0.0,0.0};

// ================================================================
// ===                                    NATALIES BIT OF SETUP                                     ===
// ================================================================

#include <SD.h>

// maximum length of the filename for an 8.3 name
#define filename_bytes 12

// chip select, not the default of 10
#define CS 6

// pin for pause_signal2()
#define pausePin 3

#define debugPin 7

// where the filename is stores
char filename[filename_bytes];

File location;
Sd2Card card;

// used so the arduino doesnt flush on every single print
int time_since_flush = 0;

#define PRINT

// ================================================================
// ===                                            INITIAL SETUP                                             ===
// ================================================================

void setup() {
  
    pinMode(debugPin, OUTPUT);
    digitalWrite(debugPin, 1);
    
    // join I2C bus (I2Cdev library doesn't do this automatically)
#if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
    Wire.begin();
    Wire.setClock(400000); // 400kHz I2C clock. Comment this line if having compilation difficulties
#endif

    // initialize serial communication
    #ifdef PRINT
    Serial.begin(115200);
    Serial.print(F("hello world"));
    #endif

    // initialize MPU
    mpu.initialize();
    pinMode(INTERRUPT_PIN, INPUT);
    devStatus = mpu.dmpInitialize();
    
    // stops the MPU from drifting
    mpu.setXGyroOffset(17);
    mpu.setYGyroOffset(-69);
    mpu.setZGyroOffset(27);
    mpu.setZAccelOffset(1551); // 1688 factory default for my test chip

    // make sure MPU workes (returns 0 if so)
    if (devStatus == 0) {
        // turn on the DMP, now that it's ready
        mpu.setDMPEnabled(true);

        // sets up an ISR on the right pin
        attachInterrupt(digitalPinToInterrupt(INTERRUPT_PIN), dmpDataReady, RISING);
        mpuIntStatus = mpu.getIntStatus();

        // set our DMP Ready flag so the main loop() function knows it's okay to use it
        dmpReady = true;

        // get expected DMP packet size for later comparison
        packetSize = mpu.dmpGetFIFOPacketSize();
    } else {
        // ERROR!
        // 1 = initial memory load failed
        // 2 = DMP configuration updates failed
        #ifdef PRINT
         Serial.print(F("DMP Initialization failed (code "));
        Serial.print(devStatus);
        Serial.println(F(")"));
        #endif
    }
    #ifdef PRINT
    Serial.println(F("mpu"));
    #endif
    
    // Define the pins to which the 3 servo motors are connected
    // this is done experimentally
    servo0.attach(10); 
    servo1.attach(9); 
    servo2.attach(8); 


// ================================================================
// ===                                         SD CARD STUFF                                                ===
// ================================================================


    pinMode(CS, OUTPUT);
    pinMode(pausePin, INPUT);

    // used for pause_signal()
    pinMode(A0, INPUT);
 
    // will almost definitely lead to a fifo overflow
    init_card(CS);
    #ifdef PRINT
    Serial.println(F("done with init"));
    #endif

    // flash debug LED to say the setup step is done
    digitalWrite(debugPin, 0);
    delay(100);
    digitalWrite(debugPin, 1);
    delay(100);
    digitalWrite(debugPin, 0);
}
// ================================================================
// ===                                        MAIN PROGRAM LOOP                                         ===
// ================================================================

void loop() {
    #ifdef PRINT       
      Serial.println(F("new loop started"));
    #endif

  
    // if programming failed, don't try to do anything
    if (!dmpReady) {
      #ifdef PRINT
      Serial.println(F("re upload code"));
      #endif
      return;
    }

    #ifdef PRINT
      Serial.println(F("dmp ready"));
    #endif
    
    // wait for MPU interrupt or extra packet(s) available
    while (!mpuInterrupt && fifoCount < packetSize) {
        if (mpuInterrupt && fifoCount < packetSize) {
            #ifdef PRINT
            Serial.println(F("waiting"));
            #endif
            // try to get out o33f the infinite loop
            fifoCount = mpu.getFIFOCount();
        }
    }

  #ifdef PRINT
            Serial.println(F("packets ready"));
            #endif

    // reset interrupt flag and get INT_STATUS byte
    mpuInterrupt = false;
    mpuIntStatus = mpu.getIntStatus();

    // get current FIFO count
    fifoCount = mpu.getFIFOCount();

    // check for overflow (this should never happen unless our code is too inefficient)
    if ((mpuIntStatus & _BV(MPU6050_INTERRUPT_FIFO_OFLOW_BIT)) || fifoCount >= 128) {
        // reset so we can continue cleanly
        mpu.resetFIFO();
        fifoCount = mpu.getFIFOCount();
        #ifdef PRINT
        Serial.println(F("FIFO overflow!"));
        #endif

        // clear FIFO buffer so there's no risk of aliasing
        for (int i = 0; i < 64; i++){
          fifoBuffer[i] = 0;
        }
        

        // otherwise, check for DMP data ready interrupt (this should happen frequently)
    } else if (mpuIntStatus & _BV(MPU6050_INTERRUPT_DMP_INT_BIT)) {
        // wait for correct available data length, should be a VERY short wait
        while (fifoCount < packetSize) fifoCount = mpu.getFIFOCount();
        #ifdef PRINT
            Serial.println(F("fifo<packet"));
        #endif
        // read a packet from FIFO
        mpu.getFIFOBytes(fifoBuffer, packetSize);

        // track FIFO count here in case there is > 1 packet available
        // (this lets us immediately read more without waiting for an interrupt)
        fifoCount -= packetSize;

        // Get Yaw, Pitch and Roll values

        #ifdef PRINT
            Serial.println(F("doing quaternion"));
        #endif

        // does some coordinate conversions. writes to pointers instead of returning arrays
        mpu.dmpGetQuaternion(&q, fifoBuffer);
        mpu.dmpGetGravity(&gravity, &q);
        mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);
        
        // radians -> degrees so its easier to read
        for (int i = 0; i < 3; i++) {
            ypr[i] = ypr[i] * 180/M_PI;
        }

        #ifdef PRINT
            Serial.println(F("mpu stuff done"));
        #endif
        
        // Skip 300 readings (self-calibration process)
        
        if (j <= 300) {
            #ifdef PRINT
            Serial.println(j);
            #endif
            j++;
        } else {
            // branch to run after calibration
            
            // Map the values of the MPU6050 sensor from -90 to 90 to values suatable for the servo control from 0 to 180
            float servo_values[3];
            // calculate the values to send to the servos and maybeprint them
            for (int i = 0; i < 3; i++) {
                servo_values[i] = lerp(ypr[i], -90.0, 90.0, 1150.0, 1850.0) * corrections[i];
            }
            
            #ifdef PRINT
            Serial.println(F("servo done"));
            #endif
            
            // Control the servos according to the MPU6050 orientation
            

            if (pause_signal2()) {
              servo0.write(90);
              servo1.write(90);
              servo2.write(90);

              // very easy to change gimbal behaviour as desired
              for (int i = 0; i < 3; i++){
                offset[i] = servo_values[i] - 1500.0;
              }
              
            }else{
              
              // run by default
              servo0.writeMicroseconds(servo_values[0] - offset[0]);
              servo1.writeMicroseconds(servo_values[1] - offset[1]);
              servo2.writeMicroseconds(servo_values[2] - offset[2]);
            }
            #ifdef PRINT
            Serial.println(F("done checking for pause and wriing position"));
            #endif
  
            // log whatever is wanted to the SD card. in this case, it is the orientation of the MPU and the servo values
            float now = millis();
            float buf[6] = {ypr[0], ypr[1], ypr[2], servo_values[0], servo_values[1], servo_values[2]};
            format_bytes(buf, 6);

            #ifdef PRINT                   
            Serial.println(F("wrote to card"));
            #endif

            
            #ifdef PRINT    
            // print values if it is enabled 
            for (int i = 0; i < 3; i++) {
                Serial.print("    ");
                Serial.print(servo_values[i]-offset[i]); 
            }
            Serial.println();
            #endif

        }
    }
    delay(12);
}

float lerp(float x, float in_min, float in_max, float out_min, float out_max) {
    // reimplementation of map for floats, will lead to smoother movement
    return max(min((x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min, out_max), out_min);
}

bool pause_signal() {
  // use this with a pause signal from the reciever
  return pulseIn(A0, HIGH) > 1600;
}

bool pause_signal2() {
  // use this for bench tests without reciever
  return digitalRead(pausePin);
}

void init_card(int pin) {
    /*
     * tries to intialise the SD card and get the filename to use
     */ 

    #ifdef PRINT
    Serial.println(F("initialising SD card"));
    Serial.println(SD.begin(pin)); // should print a 1 to the serial console if all goes well
    #endif
    
    // can the card be initialised?
    if (!card.init(SPI_HALF_SPEED, CS)) {
      #ifdef PRINT
         Serial.println(F("couldn't initialise SD card, exiting"));
         #endif
         while (true) {
            delay(1);
         }
    }
    // finds the filename to use and stores it in global memory
    get_filename();
    // finds the location to store to and saves it in global memory to write to
    location = SD.open(filename, FILE_WRITE);

    #ifdef PRINT
    Serial.print(F("activating the SD card worked, filename of "));
    Serial.println(filename);
    #endif

}

void format_bytes(float data[], int arr_size) {
    /*
      This code takes in an array of floating point numbers and iterates over it
      For each float, it reads the 4 bytes that represent it in memory and then copies them into a buffer that is used to write to the SD card
      It separates each value with a comma and then terminates the string as all c strings should be terminated
      It finally writes the bytes to the card
    */
    int len = 5*arr_size + 5;
    char *output  = (char*) malloc(len); // overestimate of the length of the output string s
    for (int i = 0; i < arr_size; i++) {
        float x = data[i];
        // type cast the bytes of the float to chars and separate with comma
        output[5 * i + 0] = ((char*)&x)[0];
        output[5 * i + 1] = ((char*)&x)[1]; 
        output[5 * i + 2] = ((char*)&x)[2];
        output[5 * i + 3] = ((char*)&x)[3];
        output[5 * i + 4] = ',';
    }
    // terminates the c string according to various standards
    output[5*arr_size] = '\0';
    write_bytes_to_card(output, len);

    // memory leaks are not ideal
    free(output);
}

void write_bytes_to_card(char* to_write, int len) {
     /*
     * takes in a filename and writes to it, making sure the file is actually written to 
     */
    
    location.write(to_write, len);
    location.write('\n');

    // increase 10 for faster code but less reliable data collection
    if (time_since_flush == 10) {
        // minimises how often the data is actually written to the card. Because flushing takes a constant, relatively long time, the less frequently the better
        
        //flash the debug LED when flushing
        digitalWrite(debugPin, 1);
        location.flush();
        time_since_flush = 0;
    } else {
        digitalWrite(debugPin, 0);
        time_since_flush++;
    }
}


void get_filename() {
    /*
     * function to get the filename to use, so each test gets a unique 
     * file
     * 
     */
    int i = 0;

    // files will follow the pattern test(number).csv
    char base_name[] = "3a";
    char ending[] = ".txt";

    
    while (true) {
        // this checks every loop to see if a filename is free. 
        // if so, it returns it
        // if not, it tries the next one
        
        char test_filename[filename_bytes];
        sprintf(test_filename, "%s%d%s", base_name, i, ending); // formats the filename
        if (SD.exists(test_filename)) {

            i++;
        }else{
            strcpy(filename, test_filename);
            break;
        }
    }
}
